<?php

namespace Code16\Sharp\Utils\Transformers;

use Closure;
use Code16\Sharp\EntityList\Commands\Command;
use Code16\Sharp\Form\Fields\Embeds\SharpFormEditorEmbed;
use Code16\Sharp\Form\SharpForm;
use Code16\Sharp\Show\SharpShow;
use Illuminate\Contracts\Pagination\LengthAwarePaginator as LengthAwarePaginatorContract;
use Illuminate\Pagination\LengthAwarePaginator;

/**
 * This trait allows a class to handle a custom transformers array.
 */
trait WithCustomTransformers
{
    protected array $transformers = [];

    public function setCustomTransformer(string $attribute, string|Closure|SharpAttributeTransformer $transformer): self
    {
        if (! $transformer instanceof SharpAttributeTransformer) {
            $transformer = $transformer instanceof Closure
                ? $this->normalizeToSharpAttributeTransformer($transformer)
                : app($transformer);
        }

        $this->transformers[$attribute] = $transformer;

        return $this;
    }

    /**
     * Transforms a model or a models collection into an array.
     */
    public function transform($models): array|LengthAwarePaginator
    {
        if ($this instanceof SharpForm || $this instanceof Command || $this instanceof SharpFormEditorEmbed) {
            // It's a Form (full entity or from a Command), there's only one model.
            // We must add Form Field Formatters in the process
            return $this->applyFormatters(
                $this->applyTransformers($models),
            );
        }

        if ($this instanceof SharpShow) {
            // It's a Show, there's only one model.
            return $this->applyTransformers($models, false);
        }

        // SharpEntityList case

        if ($models instanceof LengthAwarePaginatorContract) {
            return new LengthAwarePaginator(
                $this->transform($models->items()),
                $models->total(),
                $models->perPage(),
                $models->currentPage(),
            );
        }

        return collect($models)
            ->map(fn ($model) => $this->applyTransformers($model))
            ->all();
    }

    protected static function normalizeToSharpAttributeTransformer(Closure $closure): SharpAttributeTransformer
    {
        return new class($closure) implements SharpAttributeTransformer
        {
            public function __construct(private Closure $closure)
            {
            }

            public function apply($value, $instance = null, $attribute = null)
            {
                return call_user_func($this->closure, $value, $instance, $attribute);
            }
        };
    }

    protected function applyTransformers(array|object $model, bool $forceFullObject = true): array
    {
        $attributes = ArrayConverter::modelToArray($model);

        if ($forceFullObject) {
            // Merge model attributes with form fields to be sure we have
            // all attributes which the front code needed.
            $attributes = collect($this->getDataKeys())
                ->flip()
                ->map(fn () => null)
                ->merge($attributes)
                ->all();
        }

        if (is_object($model)) {
            $attributes = $this->handleAutoRelatedAttributes($attributes, $model);
        }

        // Apply transformers
        foreach ($this->transformers as $attribute => $transformer) {
            if (str_contains($attribute, '[')) {
                // List item case: apply transformer to each item
                $listAttribute = substr($attribute, 0, strpos($attribute, '['));
                $itemAttribute = substr($attribute, strpos($attribute, '[') + 1, -1);

                if (! array_key_exists($listAttribute, $attributes)) {
                    continue;
                }

                $listModel = $model;
                if (($sep = strpos($listAttribute, ':')) !== false) {
                    $listModel = $model[substr($listAttribute, 0, $sep)];
                    $listAttribute = substr($listAttribute, $sep + 1);
                }

                foreach ($listModel[$listAttribute] as $k => $itemModel) {
                    $attributes[$listAttribute][$k][$itemAttribute] = $transformer->apply(
                        $attributes[$listAttribute][$k][$itemAttribute] ?? null, $itemModel, $itemAttribute,
                    );
                }
            } else {
                if (! isset($attributes[$attribute])) {
                    if (method_exists($transformer, 'applyIfAttributeIsMissing')
                        && ! $transformer->applyIfAttributeIsMissing()) {
                        // The attribute is missing and the transformer code declared to be ignored in this case
                        continue;
                    }
                }

                // This is a BC, even if this is a bugfix: in case of a sub-attribute,
                // we should pass the related model to the transformer. Will do that in 9.x
//                if (($sep = strpos($attribute, ':')) !== false) {
//                    $attributes[$attribute] = $transformer->apply(
//                        $attributes[$attribute] ?? null,
//                        $model[substr($attribute, 0, $sep)] ?? null,
//                        substr($attribute, $sep + 1),
//                    );
//
//                    continue;
//                }

                $attributes[$attribute] = $transformer->apply(
                    $attributes[$attribute] ?? null,
                    $model,
                    $attribute,
                );
            }
        }

        return $attributes;
    }

    protected function applyFormatters(array $attributes): array
    {
        return collect($attributes)
            ->map(function ($value, $key) {
                $field = $this->findFieldByKey($key);

                return $field
                    ? $field->formatter()->toFront($field, $value)
                    : $value;
            })
            ->all();
    }

    /**
     * Handle `:` separator: we want to transform a related attribute in
     * a hasOne or belongsTo relationship. Ex: with "mother:name",
     * we add a transformed mother:name attribute in the array.
     */
    protected function handleAutoRelatedAttributes(array $attributes, $model): array
    {
        collect($this->getDataKeys())
            ->filter(fn (string $key) => str_contains($key, ':'))
            ->map(fn (string $key) => array_merge([$key], explode(':', $key)))
            ->each(function (array $key) use (&$attributes, $model) {
                // For each one, we create a "relation:attribute" key in the returned array
                $attributes[$key[0]] = $model->{$key[1]}->{$key[2]} ?? null;
            });

        return $attributes;
    }
}
